# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from flwr.proto import transport_pb2 as flwr_dot_proto_dot_transport__pb2


class FlowerServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.Join = channel.stream_stream(
        '/flower.transport.FlowerService/Join',
        request_serializer=flwr_dot_proto_dot_transport__pb2.ClientMessage.SerializeToString,
        response_deserializer=flwr_dot_proto_dot_transport__pb2.ServerMessage.FromString,
        )


class FlowerServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def Join(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_FlowerServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'Join': grpc.stream_stream_rpc_method_handler(
          servicer.Join,
          request_deserializer=flwr_dot_proto_dot_transport__pb2.ClientMessage.FromString,
          response_serializer=flwr_dot_proto_dot_transport__pb2.ServerMessage.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'flower.transport.FlowerService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))
